/*
 * yasapi_player.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#include <avrt.h>

///////////////////////////////////////////////////////////////////////////////
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd316756%28v=vs.85%29.aspx
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd370844%28v=vs.85%29.aspx

///////////////////////////////////////////////////////////////////////////////
#define PLAYER_FRAMES_MS(nFrames,pwfx) \
  MulDiv(nFrames,1000,(pwfx)->nSamplesPerSec)
#define PLAYER_FRAMES_TIMER(nFrames,pwfx) \
  (10000000.0*(nFrames)/(pwfx)->nSamplesPerSec)

#define PLAYER_MIN_BUFSIZE 8192

///////////////////////////////////////////////////////////////////////////////
int PlayerCreate(Player *pPlayer, Plugin *pPlugin, int samplerate,
    int numchannels, int bitspersamp)
{
  Options *pOptions=pPlugin->pOptions;
#if defined (YASAPI_DEBUG) // {
  int debug=pOptions->debug;
#else // } {
  int debug=0;
#endif // }
  AUDCLNT_SHAREMODE eShareMode=pOptions->eShareMode;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
  IMMDeviceEnumerator *pEnumerator=pPlugin->pEnumerator;
  IMMDevice *pDevice=pPlugin->pDevice;
  HRESULT hr;
  UINT32 nBlockAlign;
  IAudioClient *pAudioClient;
  IAudioRenderClient *pRenderClient;
  WAVEFORMATEX *pClosestMatch,**ppClosestMatch;
  REFERENCE_TIME hnsDevicePeriod;
  UINT32 nBufSize;
  SyncConfig sc;

  if (0==pPlugin->initialized) {
    DMESSAGE(debug,"not initialized");
    goto initialized;
  }

  pPlayer->strategy=pOptions->pull?PlayerPull():PlayerPush();
  DPRINTF(debug,"  initializing for %s mode\n",pPlayer->strategy->name);
  pPlayer->pPlugin=pPlugin;

  // setup WAVEFORMATEXTENSIBLE.
  WaveFormatExtensibleSetup(pPlayer,samplerate,numchannels,bitspersamp);

  // from the device, get an audio client.
  hr=pDevice->lpVtbl->Activate(pDevice,
    &IID_IAudioClient,        // [in]   REFIID iid,
    CLSCTX_ALL,               // [in]   DWORD dwClsCtx,
    NULL,                     // [in]   PROPVARIANT *pActivationParams,
    (void **)&pAudioClient    // [out]  void **ppInterface
  );

  if (FAILED(hr)) {
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DMESSAGE(debug,"activating audio client");
    goto client;
  }

  pPlayer->pAudioClient=pAudioClient;
  DPUTS(debug,"  audio client activated\n");

  // check whether the format is supported.
  pClosestMatch=NULL;
  ppClosestMatch=AUDCLNT_SHAREMODE_EXCLUSIVE==eShareMode?NULL:&pClosestMatch;

  hr=pAudioClient->lpVtbl->IsFormatSupported(pAudioClient,
    eShareMode,             // [in]   AUDCLNT_SHAREMODE ShareMode,
    pwfx,                   // [in]   const WAVEFORMATEX *pFormat,
    ppClosestMatch          // [out]  WAVEFORMATEX **ppClosestMatch
  );

  if (FAILED(hr)||S_FALSE==hr) {
    DERROR(debug,S_FALSE,hr);
    DERROR(debug,AUDCLNT_E_UNSUPPORTED_FORMAT,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGEV(debug,"format not supported (channels: %d, rate: %d, bps: %d)",
        pwfx->nChannels,pwfx->nSamplesPerSec,pwfx->wBitsPerSample);

    if (NULL!=pClosestMatch) {
      DPRINTF(debug,"  closest match: channels: %d, rate: %d, bps: %d\n",
          pClosestMatch->nChannels,pClosestMatch->nSamplesPerSec,
          pClosestMatch->wBitsPerSample);
      CoTaskMemFree(pClosestMatch);
    }

    goto support;
  }

  DPRINTF(debug,"  format supported (channels: %d, rate: %d, bps: %d)\n",
      pwfx->nChannels,pwfx->nSamplesPerSec,pwfx->wBitsPerSample);

  // get the device period.
#if 1 // {
  // get the minimum device period (low latency).
  hr=pAudioClient->lpVtbl->GetDevicePeriod(pAudioClient,
    NULL,                 // [out]  REFERENCE_TIME *phnsDefaultDevicePeriod,
    &hnsDevicePeriod      // [out]  REFERENCE_TIME *phnsMinimumDevicePeriod
  );
#else // } {
  // get the default device period.
  hr=pAudioClient->lpVtbl->GetDevicePeriod(pAudioClient,
    &hnsDevicePeriod,     // [out]  REFERENCE_TIME *phnsDefaultDevicePeriod,
    NULL                  // [out]  REFERENCE_TIME *phnsMinimumDevicePeriod
  );
#endif // }

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"getting device period");
    goto period;
  }

  DPRINTF(debug,"  got device period (device period: %d)\n",hnsDevicePeriod);

  if (1.0<pOptions->qShareSize) {
    hnsDevicePeriod=pOptions->qShareSize*hnsDevicePeriod+0.5;
    DPRINTF(debug,"  device period scaled by %.02f (device period: %d)\n",
        pOptions->qShareSize,hnsDevicePeriod);
  }

  hr=pPlayer->strategy->InitializeAudioClient(pPlayer,hnsDevicePeriod);

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_ALREADY_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(debug,AUDCLNT_E_CPUUSAGE_EXCEEDED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_IN_USE,hr);
    DERROR(debug,AUDCLNT_E_ENDPOINT_CREATE_FAILED,hr);
    DERROR(debug,AUDCLNT_E_INVALID_DEVICE_PERIOD,hr);
    DERROR(debug,AUDCLNT_E_UNSUPPORTED_FORMAT,hr);
    DERROR(debug,AUDCLNT_E_EXCLUSIVE_MODE_NOT_ALLOWED,hr);
    DERROR(debug,AUDCLNT_E_BUFDURATION_PERIOD_NOT_EQUAL,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGEV(debug,"initializing audio client (mode: %s)",
        YASAPI_SHAREMODE_LABEL(eShareMode));
    goto initialize;
  }

  DPRINTF(debug,"  audio client initialized (mode: %s)\n",
      YASAPI_SHAREMODE_LABEL(eShareMode));

  if (pPlayer->strategy->CreateEvent(pPlayer)<0)
    goto event;

  // get the buffer size.
  //
  // This method retrieves the length of the endpoint buffer shared
  // between the client application and the audio engine. The length
  // is expressed as the number of audio frames the buffer can hold.
  // The size in bytes of an audio frame is calculated as the number
  // of channels in the stream multiplied by the sample size per channel.
  // For example, the frame size is four bytes for a stereo (2-channel)
  // stream with 16-bit samples.
  hr=pAudioClient->lpVtbl->GetBufferSize(pAudioClient,
    &pPlayer->nFrames         // [out]  UINT32 *pNumBufferFrames
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"getting buffer size");
    goto size;
  }

  DPRINTF(debug,"  got buffer size (frames: %d)\n",pPlayer->nFrames);

  // get the render client.
  hr=pAudioClient->lpVtbl->GetService(pAudioClient,
    &IID_IAudioRenderClient,  // [in]   REFIID riid,
    (void**)&pRenderClient    // [out]  void **ppv
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGE(debug,"getting render audio client");
    goto render;
  }

  pPlayer->pRenderClient=pRenderClient;
  DPUTS(debug,"  got render audio client\n");

  // create the sync.
#if defined (YASAPI_DEBUG) // {
  sc.debug=debug;
#endif // }
  sc.pClient=pPlayer;
  sc.end.state=PLAYER_END;
  sc.read.state=PLAYER_READ;
  sc.read.Condition=PlayerCanRead;
  sc.write.state=PLAYER_WRITE;
  sc.write.Condition=PlayerCanWrite;

  if (SyncCreate(&pPlayer->sync,&sc)<0) {
    DMESSAGE(debug,"creating sync");
    goto sync;
  }

  DPUTS(debug,"  sync created\n");

  // create the ring buffer.
  if (1.0<pOptions->qRingSize) {
    nBufSize=pOptions->qRingSize*pPlayer->nFrames*pwfx->nBlockAlign+0.5;
    nBufSize-=nBufSize%pwfx->nBlockAlign;
  }
  else
    nBufSize=pPlayer->nFrames*pwfx->nBlockAlign;

  if (nBufSize<PLAYER_MIN_BUFSIZE) {
    nBufSize=PLAYER_MIN_BUFSIZE;

    if (0<nBufSize%pwfx->nBlockAlign) {
      nBufSize+=pwfx->nBlockAlign;
      nBufSize-=nBufSize%pwfx->nBlockAlign;
    }
  }

  if (RingCreate(&pPlayer->ring,nBufSize,debug)<0) {
    DMESSAGEV(debug,"creating ring buffer (scale %.1f, size: %d)",
        pOptions->qRingSize,nBufSize);
    goto ring;
  }

  pPlayer->ring.Copy=PlayerCopyMemory;
  pPlayer->ring.pClient=pPlayer;

  DPRINTF(debug,"  ring buffer created (scale %.1f, size: %d)\n",
      (double)nBufSize/(pwfx->nBlockAlign*pPlayer->nFrames)+0.5,nBufSize);

  // create the render audio thread.
  pPlayer->hThread=CreateThread(
    NULL,           // _In_opt_   LPSECURITY_ATTRIBUTES lpThreadAttributes,
    0,              // _In_       SIZE_T dwStackSize,
    PlayerThread,   // _In_       LPTHREAD_START_ROUTINE lpStartAddress,
    pPlayer,        // _In_opt_   LPVOID lpParameter,
    0,              // _In_       DWORD dwCreationFlags,
    NULL            // _Out_opt_  LPDWORD lpThreadId
  );

  if (NULL==pPlayer->hThread) {
    DMESSAGE(debug,"creating render audio thread");
    goto thread;
  }

  DPUTS(debug,"  render audio thread created\n");

  return 0;
// cleanup:
thread:
  RingDestroy(&pPlayer->ring);
ring:
  SyncDestroy(&pPlayer->sync);
sync:
  pRenderClient->lpVtbl->Release(pRenderClient);
render:
size:
  pPlayer->strategy->DestroyEvent(pPlayer);
event:
initialize:
period:
support:
  pAudioClient->lpVtbl->Release(pAudioClient);
client:
initialized:
  return -1;
}

void PlayerDestroy(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  Sync *pSync=&pPlayer->sync;
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  IAudioRenderClient *pRenderClient=pPlayer->pRenderClient;

  WaitForSingleObject(pSync->hMutex,INFINITE);
  pSync->state|=PLAYER_END;
  SetEvent(pSync->read.hEvent);
  SetEvent(pSync->write.hEvent);
  ReleaseMutex(pSync->hMutex);

  WaitForSingleObject(pPlayer->hThread,INFINITE);
  CloseHandle(pPlayer->hThread);
  DPUTS(debug,"  render audio thread destroyed\n");

  RingDestroy(&pPlayer->ring);
  DPUTS(debug,"  ring buffer destroyed\n");
  SyncDestroy(&pPlayer->sync);
  DPUTS(debug,"  sync destroyed\n");

  pPlayer->strategy->DestroyEvent(pPlayer);

  pRenderClient->lpVtbl->Release(pRenderClient);
  DPUTS(debug,"  render audio client destroyed\n");
  pAudioClient->lpVtbl->Release(pAudioClient);
  DPUTS(debug,"  audio client destroyed\n");
}

///////////////////////////////////////////////////////////////////////////////
HRESULT PlayerStart(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  Sync *pSync=&pPlayer->sync;
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr=S_OK;

  WaitForSingleObject(pSync->hMutex,INFINITE);

  if (0==((PLAYER_STARTED|PLAYER_END)&pSync->state)) {
    // start the audio client.
    hr=pAudioClient->lpVtbl->Start(pAudioClient);

    if (FAILED(hr)) {
      DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
      DERROR(debug,AUDCLNT_E_NOT_STOPPED,hr);
      DERROR(debug,AUDCLNT_E_EVENTHANDLE_NOT_SET,hr);
      DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
      DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
      DMESSAGE(debug,"starting audio client");
      goto start;
    }

    pSync->state|=PLAYER_STARTED;
    DPUTS(debug,"  audio client started\n");
  }
start:
  ReleaseMutex(pSync->hMutex);
  return hr;
}

HRESULT PlayerStopAtom(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  Sync *pSync=&pPlayer->sync;
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr=S_OK;

  if (PLAYER_IS_STARTED(pSync)) {
    // start the audio client.
    hr=pAudioClient->lpVtbl->Stop(pAudioClient);

    if (FAILED(hr)) {
      DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
      DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING ,hr);
      DMESSAGE(debug,"stopping audio client");
      goto start;
    }

    DPUTS(debug,"  audio client stopped\n");
    pSync->state&=~PLAYER_STARTED;
  }
start:
  return hr;
}

HRESULT PlayerStop(Player *pPlayer)
{
  HANDLE hMutex=pPlayer->sync.hMutex;
  HRESULT hr;

  WaitForSingleObject(hMutex,INFINITE);
  hr=PlayerStopAtom(pPlayer);
  ReleaseMutex(hMutex);

  return hr;
}

SIZE_T PlayerAvailable(Player *pPlayer)
{
  return pPlayer->ring.dwAvailable;
}

BOOL PlayerCanWrite(LPVOID pClient)
{
  return 0<((Player *)pClient)->ring.dwAvailable;
}

BOOL PlayerCanRead(LPVOID pClient)
{
  Player *pPlayer=pClient;
  Sync *pSync=&pPlayer->sync;

  return !PLAYER_IS_PAUSE(pSync)&&pPlayer->strategy->CanRead(pPlayer);
}

////////
SIZE_T PlayerWrite(Player *pPlayer, LPSTR pData, SIZE_T dwSize)
{
  Sync *pSync=&pPlayer->sync;

  dwSize-=dwSize%pPlayer->wfxx.Format.nBlockAlign;
  SyncWaitWrite(pSync);
  dwSize=RingWrite(&pPlayer->ring,pData,dwSize);
  SyncReleaseWrite(pSync);

  return dwSize;
}

HRESULT PlayerResetAtom(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr;

  hr=pAudioClient->lpVtbl->Reset(pAudioClient);

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_NOT_STOPPED,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_OPERATION_PENDING,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGE(debug,"resetting audio client");
    goto reset;
  }
reset:
  return hr;
}

void PlayerReset(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  Sync *pSync=&pPlayer->sync;
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr;

  WaitForSingleObject(pSync->hMutex,INFINITE);

  if (PLAYER_IS_STARTED(pSync)) {
    hr=PlayerStopAtom(pPlayer);

    if (FAILED(hr)) {
      DMESSAGE(debug,"stopping audio client");
      goto stop;
    }
  }

  hr=PlayerResetAtom(pPlayer);

  if (FAILED(hr)) {
    DMESSAGE(debug,"resetting audio client");
    goto reset;
  }

  RingReset(&pPlayer->ring);
reset:
stop:
  ReleaseMutex(pSync->hMutex);
}

void PlayerPause(Player *pPlayer, BOOL bPause)
{
  Sync *pSync=&pPlayer->sync;

  WaitForSingleObject(pSync->hMutex,INFINITE);

  if (0==bPause)
    pSync->state&=~PLAYER_PAUSE;
  else {
    pSync->state|=PLAYER_PAUSE;
    PlayerStopAtom(pPlayer);
  }

  SetEvent(pSync->read.hEvent);
  SetEvent(pSync->write.hEvent);
  ReleaseMutex(pSync->hMutex);
}

///////////////////////////////////////////////////////////////////////////////
HRESULT PlayerRenderFrames(Player *pPlayer, UINT32 *pNumFramesRender)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  Sync *pSync=&pPlayer->sync;
  Ring *pRing=&pPlayer->ring;
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  IAudioRenderClient *pRenderClient=pPlayer->pRenderClient;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
  UINT32 nFrames=pPlayer->nFrames;
  UINT32 nFramesPadding;
  UINT32 nFramesWrite;
  BYTE *pData;
  HRESULT hr;

  // lock the ring buffer.
  SyncWaitRead(pSync);

  if (PLAYER_IS_PAUSE(pSync)) {
    // if we are going here and are in "pause" we're in "end" too.
    if (!PLAYER_IS_END(pSync))
      DMESSAGE(debug,"inconsistent state");

    RingReset(&pPlayer->ring);
    *pNumFramesRender=0;
  }
  else {
    hr=pPlayer->strategy->GetFramesWrite(pPlayer,&nFramesWrite,
        &nFramesPadding);

    if (FAILED(hr))
      goto frames;

    // get the render buffer.
    if (0<nFramesWrite) {
      hr=pRenderClient->lpVtbl->GetBuffer(pRenderClient,
        nFramesWrite,       // [in]   UINT32 NumFramesRequested,
        &pData              // [out]  BYTE **ppData
      );

      if (FAILED(hr)) {
        DERROR(debug,AUDCLNT_E_BUFFER_ERROR,hr);
        DERROR(debug,AUDCLNT_E_BUFFER_TOO_LARGE,hr);
        DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
        DERROR(debug,AUDCLNT_E_OUT_OF_ORDER,hr);
        DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
        DERROR(debug,AUDCLNT_E_BUFFER_OPERATION_PENDING,hr);
        DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
        DERROR(debug,E_POINTER,hr);
        DMESSAGE(debug,"getting render buffer");
        goto get;
      };

//DPRINTF(debug," %d -> ",pRing->dwWritten);
      // transfer the audio data from the ring buffer into the render buffer.
      RingRead(pRing,pData,nFramesWrite*pwfx->nBlockAlign);
//DPRINTF(debug,"%d\n",pRing->dwWritten);

      // release the render buffer.
      hr=pRenderClient->lpVtbl->ReleaseBuffer(pRenderClient,
        nFramesWrite,     // [in]  UINT32 NumFramesWritten,
        0                 // [in]  DWORD dwFlags
      );

      if (FAILED(hr)) {
        DERROR(debug,AUDCLNT_E_INVALID_SIZE,hr);
        DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
        DERROR(debug,AUDCLNT_E_OUT_OF_ORDER,hr);
        DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
        DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
        DERROR(debug,E_INVALIDARG,hr);
        DMESSAGE(debug,"releasing render buffer");
        goto release;
      }
    }

    *pNumFramesRender=nFramesPadding+nFramesWrite;
  }
release:
get:
frames:
  // release the lock on the ring buffer.
  SyncReleaseRead(pSync);
  return hr;
}

DWORD WINAPI PlayerThread(LPVOID lpParameter)
{
  Player *pPlayer=lpParameter;
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  Options *pOptions=pPlugin->pOptions;
  Sync *pSync=&pPlayer->sync;
  Ring *pRing=&pPlayer->ring;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
  UINT32 nFrames=pPlayer->nFrames;
  LONGLONG time=0.5*PLAYER_FRAMES_TIMER(nFrames,pwfx)+0.5;
  UINT32 nFramesRender;
  HANDLE hTask;
  DWORD taskIndex;
  HRESULT hr;

  DPUTS(debug,"start playing\n");
  taskIndex=0;
  hTask=AvSetMmThreadCharacteristicsA("Pro Audio",&taskIndex);

  if (NULL==hTask) {
    DMESSAGE(debug,"setting thread cahracteristics");
    goto task;
  }

  if (1.0<pOptions->qRingFill)
    nFrames=pOptions->qRingFill*nFrames+0.5;

  while (!PLAYER_IS_END(pSync)) {
    if (PLAYER_IS_STARTED(pSync))
      pPlayer->strategy->Wait(pPlayer,time);

    hr=PlayerRenderFrames(pPlayer,&nFramesRender);

    if (FAILED(hr)) {
      DMESSAGE(debug,"rendering frames");
      continue;
    }

    // make sure that the renderer can be completely filled before starting.
    if (!PLAYER_IS_STARTED(pSync)&&nFrames<=pRing->dwWritten)
      PlayerStart(pPlayer);
  }

  if (PLAYER_IS_STARTED(pSync)) {
    if (PLAYER_IS_PAUSE(pSync))
      PlayerResetAtom(pPlayer);
    else {
      nFrames=0;

      for (;;) {
        // get number of remaining frames in render buffer.
        hr=pAudioClient->lpVtbl->GetCurrentPadding(pAudioClient,
          &nFramesRender      // [out]  UINT32 *pNumPaddingFrames
        );

        if (FAILED(hr))
          break;
        else if (nFrames==nFramesRender) {
#if defined (YASAPI_DEBUG) // {
          if (0<nFrames)
            DPRINTF(debug, "  sorry, %d frames lost ...\n",nFrames);
#endif // }
          break;
        }

        // remember the numer of frames in order to avoid an endless loop.
        nFrames=nFramesRender;

        // in case there are frames, wait until played.
        TimerSleep(&pPlugin->timer,PLAYER_FRAMES_TIMER(nFrames,pwfx)+0.5);
      }
    }

    PlayerStop(pPlayer);
  }

  DPUTS(debug,"  stop playing\n");
// cleanup:
  AvRevertMmThreadCharacteristics(hTask);
task:
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
void PlayerCopyMemory(PVOID pClient, PVOID Destination, const VOID *Source,
    SIZE_T Length)
{
  Player *pPlayer=pClient;
  Plugin *pPlugin=pPlayer->pPlugin;
  Options *pOptions=pPlugin->pOptions;
  double volume=pPlugin->volume;
  char *wp,*mp;
  const char *rp;
  int n,m;
  int max32,max;
  double q;
  int32_t i32;

  if (0==pOptions->volume||1.0==volume) {
    CopyMemory(
      Destination,              // _In_  PVOID Destination,
      Source,                   // _In_  const VOID *Source,
      Length                    // _In_  SIZE_T Length
    );
  }
  else {
    wp=Destination;
    mp=wp+Length;
    rp=Source;
    n=pPlayer->wfxx.Format.wBitsPerSample;
    m=n>>3;
    max32=~(~0<<(32-1));
    max=~(~0<<(n-1));
    q=volume*max/max32;

    while (wp<mp) {
      i32=0;

      CopyMemory(
        &i32,                   // _In_  PVOID Destination,
        rp,                     // _In_  const VOID *Source,
        m                       // _In_  SIZE_T Length
      );

      i32<<=32-n;
      i32=q*i32+0.5;

      CopyMemory(
        wp,                     // _In_  PVOID Destination,
        &i32,                   // _In_  const VOID *Source,
        m                       // _In_  SIZE_T Length
      );

      wp+=m;
      rp+=m;
    }
  }
}
