/*
 * yasapi_player_pull.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#include <avrt.h>

///////////////////////////////////////////////////////////////////////////////
static HRESULT PlayerInitializeAudioClientPull(Player *pPlayer,
    REFERENCE_TIME hnsDevicePeriod)
{
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  AUDCLNT_SHAREMODE eShareMode=pPlayer->pPlugin->pOptions->eShareMode;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;

  return pAudioClient->lpVtbl->Initialize(pAudioClient,
    eShareMode,           // [in]  AUDCLNT_SHAREMODE ShareMode,
    AUDCLNT_STREAMFLAGS_EVENTCALLBACK,
                          // [in]  DWORD StreamFlags,
    hnsDevicePeriod,      // [in]  REFERENCE_TIME hnsBufferDuration,
    hnsDevicePeriod,      // [in]  REFERENCE_TIME hnsPeriodicity,
    pwfx,                 // [in]  const WAVEFORMATEX *pFormat,
    NULL                  // [in]  LPCGUID AudioSessionGuid
  );
}

static int PlayerCreateEventPull(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#else // } {
  int debug=0;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr;

  // create event.
  pPlayer->hEvent=CreateEvent(
    NULL,         // _In_opt_  LPSECURITY_ATTRIBUTES lpEventAttributes,
    FALSE,        // _In_      BOOL bManualReset,
    FALSE,        // _In_      BOOL bInitialState,
    NULL          // _In_opt_  LPCTSTR lpName
  );

  if (NULL==pPlayer->hEvent) {
    DMESSAGE(debug,"creating event");
    goto event;
  }

  DPUTS(debug,"  event created\n");

  // set event.
  hr=pAudioClient->lpVtbl->SetEventHandle(pAudioClient,
    pPlayer->hEvent           // [in]  HANDLE eventHandle
  );

  if (FAILED(hr)) {
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,AUDCLNT_E_EVENTHANDLE_NOT_EXPECTED,hr);
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING ,hr);
    DMESSAGE(debug,"setting event to the audio client");
    goto set;
  }

  DPUTS(debug,"  event set to the audio client\n");

  return 0;
set:
  CloseHandle(pPlayer->hEvent);
event:
  return -1;
}

static void PlayerDestroyEventPull(Player *pPlayer)
{
  CloseHandle(pPlayer->hEvent);
}

static BOOL PlayerCanReadPull(Player *pPlayer)
{
  return pPlayer->nFrames*pPlayer->wfxx.Format.nBlockAlign
      <=pPlayer->ring.dwWritten;
}

static int PlayerWaitPull(Player *pPlayer, LONGLONG time)
{
  return WAIT_OBJECT_0==WaitForSingleObject(pPlayer->hEvent,2000)?0:-1;
}

HRESULT PlayerGetFramesWritePull(Player *pPlayer, UINT32 *pNumFramesWrite,
    UINT32 *pNumFramesPadding)
{
  *pNumFramesWrite=pPlayer->nFrames;
  *pNumFramesPadding=0;

  return S_OK;
}

const PlayerStrategy *PlayerPull(void)
{
  static int initialized;
  static PlayerStrategy strategy;

  if (0==initialized) {
    strategy.name="pull";
    strategy.InitializeAudioClient=PlayerInitializeAudioClientPull;
    strategy.CreateEvent=PlayerCreateEventPull;
    strategy.DestroyEvent=PlayerDestroyEventPull;
    strategy.CanRead=PlayerCanReadPull;
    strategy.Wait=PlayerWaitPull;
#if 0 // {
    strategy.GetFramesWrite=PlayerGetFramesWritePull;
#else // } {
    strategy.GetFramesWrite=PlayerGetFramesWritePush;
#endif // }
    initialized=1;
  }

  return &strategy;
}
